//! # **C** code based **Y**et **A**nother **GEN**erator
//! - Scan C source file using the simple pattern matching to capture the elements in the code
//! - Generate text based files using template files
//! - Supported elements are inclusion, local variable, and functions
//!

extern crate chrono;
extern crate regex;

use chrono::Utc;
use regex::Regex;

#[derive(Debug)]
pub struct Include {
    pub captured: String,
}

#[derive(Debug)]
pub struct LocalVariable {
    pub captured: String,
    pub name: String,
    pub dtype: String,
}

#[derive(Debug, Clone)]
pub struct Function {
    pub captured: String,
    pub name: String,
    pub is_local: bool,
    pub rtype: String,
    pub args: String,
    pub atypes: String,
}

#[derive(Debug)]
pub struct NestedCall {
    pub callee: Function,
    pub caller: Function,
}

#[derive(Debug)]
pub struct Parser {
    incs: Vec<Include>,
    local_vars: Vec<LocalVariable>,
    fncs: Vec<Function>,
    ncls: Vec<NestedCall>,
}

impl Parser {
    /// parse the given textdata and return Parse object to be used for generator
    ///
    pub fn parse<'a>(textdata: &'a str) -> Self {
        let code = remove_comments(textdata);
        let fncs = get_fncs(&code);
        let ncls = get_ncls(&code, &fncs);
        Self {
            incs: get_incs(&code),
            local_vars: get_local_vars(&code),
            fncs: fncs.clone(),
            ncls: ncls,
        }
    }
}

/// generate document based on parsing result and template data
///
/// # Available tags in template file
/// - **@sourcename@** : it is given as an argument from command line
/// - **@date@** : generated date
/// - **@incs@** : the list of inclusion statement such as `#include <stdio.h>`
///     - **@captured@** : the captured raw string
/// - **@end-incs@** : the end of **incs** block
/// - **@local-vars@** : the list of **static** variables
///     - **@captured@** : the captured raw string
///     - **@name@** : variable name
///     - **@dtype@** : variable data type
/// - **@end-local-vars@** : the end of **local-var** bolck
/// - **@fncs@** or **@fncs0@** : the list of all the functions
///     - **@captured@** : the captured raw string
///     - **@name@** : the function name
///     - **@rtype@** : the return data type of the function
///     - **@args@** : the list of arguments with data types
///     - **@atypes@** : the list of only arguments' data types
/// - **@end-fncs@** or **@end-fncs0@** : the end of **fncs** or **fncs0** block
/// - **@ncls@** or **@ncls-once@** : the list of nested calls, no duplicate callee with **ncls-once**
///     - **@callee.name@** : the function name of callee
///     - **@callee.rtype@** : the return type of callee
///     - **@callee.rtype.change(<from>=<to>)@** : to change return data type during generation
///     - **@callee.rtype.variable(<var>)@** : <var> string to be removed when `void`
///     - **@callee.rtype.value(<val>)@** : <val> string to be removed when `void`
///     - **@callee.args@** : the argument list string
///     - **@callee.atypes@** : only arguments' data types
///     - **@caller.name@** : the function name of caller
///     - **@caller.rtype@** : the return type of caller
///     - **@caller.args@** : the argument list string
///     - **@caller.atypes@** : only arguments' data types
/// - **@end-ncls@** or **@end-ncls-once@** : the end of **ncls** or **ncls-once** block
///
/// # Example
///
/// ```
/// let sourcename = "source";
/// let code = "\
/// #include <stdio.h>
/// static int var = 1;
/// static int func1(void)
/// {
///     return 0;
/// }
/// int func2(char c)
/// {
///     return func1();
/// }
/// ";
/// let temp = "\
/// // include
/// @incs@@captured@
/// @end-incs@
/// // local variables
/// @local-vars@@dtype@ @name@;
/// @end-local-vars@
/// // functions
/// @fncs@@rtype@ @name@(@args@);
/// @end-fncs@
/// ";
/// let parser = cyagen::Parser::parse(code);
/// let gen = cyagen::generate(&parser, temp, sourcename);
/// ```
pub fn generate<'a>(parser: &'a Parser, template: &'a str, sourcename: &'a str) -> String {
    let mut output = String::from(template);
    if template.contains("@incs@") {
        let re = Regex::new(r"@incs@(?P<fmt>[\S\s]*)@end-incs@").unwrap();
        for cap in re.captures_iter(template) {
            let mut tmpstr = String::new();
            for entry in &parser.incs {
                let fmtstr = cap
                    .name("fmt")
                    .unwrap()
                    .as_str()
                    .replace("@captured@", &entry.captured);
                tmpstr.push_str(&fmtstr);
            }
            output = re.replace(&output, tmpstr.as_str()).into_owned();
        }
    }
    if template.contains("@local-vars@") {
        let re = Regex::new(r"@local-vars@(?P<fmt>[\S\s]*)@end-local-vars@").unwrap();
        for cap in re.captures_iter(template) {
            let mut tmpstr = String::new();
            for entry in &parser.local_vars {
                let fmtstr = cap
                    .name("fmt")
                    .unwrap()
                    .as_str()
                    .replace("@captured@", &entry.captured)
                    .replace("@name@", &entry.name)
                    .replace("@dtype@", &entry.dtype);
                tmpstr.push_str(&fmtstr);
            }
            output = re.replace(&output, tmpstr.as_str()).into_owned();
        }
    }
    let fncs_tags = vec!["fncs", "fncs0"];
    for tag in fncs_tags {
        let regstr = format!("@{}@{}@end-{}@", tag, r"(?P<fmt>[\S\s]*)", tag);
        let re = Regex::new(&regstr).unwrap();
        for cap in re.captures_iter(template) {
            let mut tmpstr = String::new();
            for entry in &parser.fncs {
                let fmtstr = cap
                    .name("fmt")
                    .unwrap()
                    .as_str()
                    .replace("@captured@", &entry.captured)
                    .replace("@name@", &entry.name)
                    .replace("@rtype@", &entry.rtype)
                    .replace("@args@", &entry.args)
                    .replace("@atypes@", &entry.atypes);
                tmpstr.push_str(&fmtstr);
            }
            output = re.replace(&output, tmpstr.as_str()).into_owned();
        }
    }
    if output.contains("@local-fncs@") {
        let re = Regex::new(r"@local-fncs@(?P<fmt>[\S\s]*)@end-local-fncs@").unwrap();
        for cap in re.captures_iter(template) {
            let mut tmpstr = String::new();
            for entry in &parser.fncs {
                if entry.is_local {
                    let fmtstr = cap
                        .name("fmt")
                        .unwrap()
                        .as_str()
                        .replace("@captured@", &entry.captured)
                        .replace("@name@", &entry.name)
                        .replace("@rtype@", &entry.rtype)
                        .replace("@args@", &entry.args)
                        .replace("@atypes@", &entry.atypes);
                    tmpstr.push_str(&fmtstr);
                }
            }
            output = re.replace(&output, tmpstr.as_str()).into_owned();
        }
    }
    let ncls_tags = vec!["ncls", "ncls-once"];
    for tag in ncls_tags {
        let regstr = format!("@{}@{}@end-{}@", tag, r"(?P<fmt>[\S\s]*)", tag);
        let re = Regex::new(&regstr).unwrap();
        for cap in re.captures_iter(template) {
            let mut tmpstr = String::new();
            let mut callee_list: Vec<String> = Vec::new();
            for entry in &parser.ncls {
                if tag == "ncls-once" {
                    if callee_list.contains(&entry.callee.name) {
                        continue;
                    }
                    callee_list.push(entry.callee.name.to_string());
                }
                let mut fmtstr = cap
                    .name("fmt")
                    .unwrap()
                    .as_str()
                    .replace("@callee.name@", &entry.callee.name)
                    .replace("@callee.rtype@", &entry.callee.rtype)
                    .replace("@callee.args@", &entry.callee.args)
                    .replace("@callee.atypes@", &entry.callee.atypes)
                    .replace("@caller.name@", &entry.caller.name)
                    .replace("@caller.rtype@", &entry.caller.rtype)
                    .replace("@caller.args@", &entry.caller.args)
                    .replace("@caller.atypes@", &entry.caller.atypes);
                let re4change = Regex::new(
                    r"@callee.rtype.change\((?P<from>[a-z|A-Z|0-9|_]+)=(?P<to>[a-z|A-Z|0-9|_]+)\)@",
                )
                .unwrap();
                for cap in re4change.captures_iter(fmtstr.clone().as_str()) {
                    if cap.name("from").unwrap().as_str() == entry.callee.rtype.as_str() {
                        let to = cap.name("to").unwrap().as_str();
                        fmtstr = re4change.replace(&fmtstr, to).into_owned();
                    } else {
                        fmtstr = re4change.replace(&fmtstr, &entry.callee.rtype).into_owned();
                    }
                }
                let re4variable =
                    Regex::new(r"@callee.rtype.variable\((?P<variable>[a-z|A-Z|0-9|_]+)\)@")
                        .unwrap();
                for cap in re4variable.captures_iter(fmtstr.clone().as_str()) {
                    if entry.callee.rtype.as_str() == "void" {
                        fmtstr = re4variable.replace(&fmtstr, "").into_owned();
                    } else {
                        let variable = cap.name("variable").unwrap().as_str();
                        fmtstr = re4variable.replace(&fmtstr, variable).into_owned();
                    }
                }
                let re4value =
                    Regex::new(r"@callee.rtype.value\((?P<value>[a-z|A-Z|0-9|_]+)\)@").unwrap();
                for cap in re4value.captures_iter(fmtstr.clone().as_str()) {
                    if entry.callee.rtype.as_str() == "void" {
                        fmtstr = re4value.replace(&fmtstr, "").into_owned();
                    } else {
                        let value = cap.name("value").unwrap().as_str();
                        fmtstr = re4value.replace(&fmtstr, value).into_owned();
                    }
                }
                tmpstr.push_str(&fmtstr);
            }
            output = re.replace(&output, tmpstr.as_str()).into_owned();
        }
    }
    output.replace("@sourcename@", sourcename).replace(
        "@date@",
        Utc::now().format("%a %b %e %T %Y").to_string().as_str(),
    )
}

/// remove comments from C source code
///
fn remove_comments<'a>(code: &'a str) -> String {
    let re = Regex::new(r"(/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/)|(//.*)").unwrap();
    re.replace_all(&code, "").to_string()
}

/// list of inclusion from C source code
///
fn get_incs<'a>(code: &'a str) -> Vec<Include> {
    let mut result = vec![];
    let re = Regex::new(r#"[\n]{0,1}[\t\s]*(?P<captured>#include[\t\s]+["<].+[">])"#).unwrap();
    for cap in re.captures_iter(code) {
        result.push(Include {
            captured: cap.name("captured").unwrap().as_str().trim().to_string(),
        });
    }
    result
}

/// list of local variables from C source code
///
fn get_local_vars<'a>(code: &'a str) -> Vec<LocalVariable> {
    let mut result = vec![];
    let re =
        Regex::new(r"[\n]{0,1}[\t\s]*(?i)(static)[\t\s\n]*(?P<captured>[^\(\{;=]+)[;=]").unwrap();
    for cap in re.captures_iter(code) {
        let name = cap
            .name("captured")
            .unwrap()
            .as_str()
            .trim()
            .split(' ')
            .collect::<Vec<&str>>()
            .last()
            .unwrap()
            .trim_matches('*')
            .to_string();
        let data_type = cap
            .name("captured")
            .unwrap()
            .as_str()
            .trim()
            .strip_suffix(&name)
            .unwrap()
            .trim()
            .to_string();
        result.push(LocalVariable {
            captured: cap.name("captured").unwrap().as_str().trim().to_string(),
            name: name,
            dtype: data_type,
        });
    }
    result
}

/// list of functions from C source code
///
fn get_fncs<'a>(code: &'a str) -> Vec<Function> {
    let mut result = vec![];
    let re = Regex::new(
        r"[\n][\s\t]*(?P<return_func>([a-zA-Z_][a-zA-Z_0-9]*\s+)+([a-zA-Z_][a-zA-Z_0-9]*\s*))\((?P<args>[^=!><>;\(\)-]*)\)[\s\t\r\n]*\{",
    ).unwrap();
    let get_types = |args: &'a str| -> String {
        let mut type_list = String::new();
        let mut first_pos = true;
        let arg_list = args.split(',').collect::<Vec<&str>>();
        for arg in arg_list {
            let arg = arg.trim();
            if let Some(pos) = arg.rfind(|c: char| (c == '*') || (c == ' ')) {
                if first_pos {
                    first_pos = false;
                } else {
                    type_list.push_str(", ");
                }
                let mut tmpstr = arg.get(..(pos + 1)).unwrap().trim().to_string();
                let re4const = Regex::new(r"\w[\s\r\n]+const[\s\r\n]*\*").unwrap();
                if let Some(_) = re4const.captures(&tmpstr) {
                    let pos = tmpstr.find("const").unwrap();
                    tmpstr.replace_range(pos..(pos + 5), "");
                    tmpstr = format!("const {}", tmpstr);
                }
                type_list.push_str(&tmpstr);
                let array_dimension = arg.get(pos..).unwrap().matches("[").count();
                type_list.push_str(&"*".repeat(array_dimension));
            }
        }
        if type_list.trim() == "void" {
            type_list.clear();
        }
        type_list
    };
    for cap in re.captures_iter(code) {
        let name = cap
            .name("return_func")
            .unwrap()
            .as_str()
            .trim()
            .split(' ')
            .collect::<Vec<&str>>()
            .last()
            .unwrap()
            .trim_matches('*')
            .to_string();
        let return_type = cap
            .name("return_func")
            .unwrap()
            .as_str()
            .trim()
            .strip_suffix(&name)
            .unwrap()
            .replace("static", "")
            .replace("STATIC", "")
            .replace("inline", "")
            .replace("INLINE", "")
            .trim()
            .to_string();
        let is_local = cap
            .name("return_func")
            .unwrap()
            .as_str()
            .contains("static ")
            || cap
                .name("return_func")
                .unwrap()
                .as_str()
                .contains("STATIC ");
        let re4space = Regex::new(r"\s+").unwrap();
        let mut args = cap
            .name("args")
            .unwrap()
            .as_str()
            .trim()
            .to_string()
            .replace("\n", "")
            .replace("\r", "");
        args = re4space.replace_all(&args, " ").to_string();
        let mut args_type = get_types(cap.name("args").unwrap().as_str());
        args_type = re4space.replace_all(&args_type, " ").to_string();
        result.push(Function {
            captured: cap.get(0).unwrap().as_str().trim().to_string(),
            name: name,
            is_local: is_local,
            rtype: return_type,
            args: args,
            atypes: args_type,
        });
    }
    result
}

/// list of ncls in C source
///
fn get_ncls<'a>(code: &'a str, fncs: &Vec<Function>) -> Vec<NestedCall> {
    let mut result = vec![];
    for caller in fncs {
        let find_end_of_func = |start: usize| -> usize {
            let mut stop = start;
            let mut open = 1;
            for (i, c) in code.get(start..).unwrap().chars().enumerate() {
                if c == '}' {
                    open -= 1;
                } else if c == '{' {
                    open += 1;
                }
                if open == 0 {
                    stop += i;
                    break;
                }
            }
            stop
        };
        if let Some(pos) = code.find(caller.captured.as_str()) {
            let start = pos + code.get(pos..).unwrap().find('{').unwrap() + 1;
            let stop = find_end_of_func(start);
            let body = code.get(start..stop).unwrap();
            for callee in fncs {
                let call_str = format!("{}(", callee.name);
                if body.contains(call_str.as_str()) {
                    result.push(NestedCall {
                        callee: callee.clone(),
                        caller: caller.clone(),
                    });
                }
            }
        }
    }
    result
}

#[cfg(test)]
mod tests {
    use super::*;

    static TEST_CODE: &'static str = "\
#include <stdio.h>
#include \"test.h\"

int global_var = 2;
static char local_var;

// test-comment1
/* test-comment2 */
/* 
    test-comment3
*/

static inline char local_function(int a);

void main()
{
    char c = local_function(20);
    if (c == 1)
    {
        printf(\"no operation\");
    }
    else
    {
        printf(\"hello world! %c\n\", c);
    }
}

static inline char local_function(int a, 
    int*b )
{
    return (char)a;
}
";

    #[test]
    fn test_remove_comments() {
        let clean_code = remove_comments(TEST_CODE);
        assert!(!clean_code.contains("test-comment"));
    }

    #[test]
    fn test_get_incs() {
        let list_incs = get_incs(TEST_CODE);
        assert_eq!(list_incs[0].captured, "#include <stdio.h>");
        assert_eq!(list_incs[1].captured, "#include \"test.h\"");
    }

    #[test]
    fn test_get_local_vars() {
        let list_local_vars = get_local_vars(TEST_CODE);
        assert_eq!(list_local_vars[0].name, "local_var");
        assert_eq!(list_local_vars[0].dtype, "char");
    }

    #[test]
    fn test_get_fncs() {
        let list_fncs = get_fncs(TEST_CODE);
        assert_eq!(list_fncs[0].name, "main");
        assert_eq!(list_fncs[0].rtype, "void");
        assert!(!list_fncs[0].is_local);
        assert_eq!(list_fncs[1].name, "local_function");
        assert_eq!(list_fncs[1].rtype, "char");
        assert_eq!(list_fncs[1].atypes, "int, int*");
        assert!(list_fncs[1].is_local);
    }

    #[test]
    fn test_get_ncls() {
        let list_fncs = get_fncs(TEST_CODE);
        let list_ncls = get_ncls(TEST_CODE, &list_fncs);
        if list_ncls.len() > 0 {
            assert_eq!(list_ncls[0].caller.name, "main");
            assert_eq!(list_ncls[0].callee.name, "local_function");
        } else {
            assert!(false);
        }
    }

    #[test]
    fn test_generate_incs() {
        let sourcename = "test";
        let code = "\
#include <header1.h>
#include <header2.h>
";
        let temp = "\
// include
@incs@@captured@
@end-incs@
";
        let expected = "\
// include
#include <header1.h>
#include <header2.h>

";
        let parser = Parser::parse(code);
        let generated = generate(&parser, temp, sourcename);
        assert_eq!(generated, expected);
    }

    #[test]
    fn test_generate_localvars() {
        let sourcename = "test";
        let code = "\
static int a;
int b;
static char *c;
";
        let temp = "\
// static variables
@local-vars@@dtype@ @name@;
@end-local-vars@
";
        let expected = "\
// static variables
int a;
char * c;

";
        let parser = Parser::parse(code);
        let generated = generate(&parser, temp, sourcename);
        assert_eq!(generated, expected);
    }

    #[test]
    fn test_generate_fncs() {
        let sourcename = "test";
        let code = "\
// functions
int func1()
{
    return 0;
}
void func2(int a)
{
}
";
        let temp = "\
// functions
@fncs@@rtype@ @name@(@args@);
@end-fncs@
";
        let expected = "\
// functions
int func1();
void func2(int a);

";
        let parser = Parser::parse(code);
        let generated = generate(&parser, temp, sourcename);
        assert_eq!(generated, expected);
    }

    #[test]
    fn test_generate_ncls() {
        let sourcename = "test";
        let code = "\
// functions
int func1()
{
    return 0;
}
void func2(int a)
{
    func1();
}
";
        let temp = "\
@ncls@- @caller.name@ -> @callee.name@
@end-ncls@
";
        let expected = "\
- func2 -> func1

";
        let parser = Parser::parse(code);
        let generated = generate(&parser, temp, sourcename);
        assert_eq!(generated, expected);
    }

    #[test]
    fn test_generate_ncls_once() {
        let sourcename = "test";
        let code = "\
// functions
int func1()
{
    return 0;
}
void func2(int a)
{
    func1();
}
void func3(int a)
{
    func1();
}
";
        let temp = "\
@ncls-once@- @callee.name@
@end-ncls-once@
";
        let expected = "\
- func1

";
        let parser = Parser::parse(code);
        let generated = generate(&parser, temp, sourcename);
        assert_eq!(generated, expected);
    }
}
